package POWER::Event::Epoll;

use version; $VERSION = qv('0.2.0');    # also in: POD, Changes

use warnings;
use strict;
use Carp;

use Errno qw( ENOENT );
use Scalar::Util qw( weaken );

use Perl6::Export::Attrs;

use POWER::Epoll 2.0.0;


my $EVENTMASK   = EPOLLET|EPOLLIN|EPOLLOUT|EPOLLHUP|EPOLLERR;
my $EPFD        = epoll_create(100);

my @Disp_for;   # Dispatcher OBJECT|CLASS for FILENO.
my @Events;

sub _DUMP { # DEBUG
    warn "<<POWER::Event::Epoll>>\n";
    for (grep {$Disp_for[$_]} 0 .. $#Disp_for) {
        warn sprintf "\tfd=%s\tdisp=%s\tfh=%s\n",
            $_,
            (defined $Disp_for[$_][0] 
                ? $Disp_for[$_][0]
                : '(destroyed object)'
            ),
            $Disp_for[$_][1],
            ;
    }
}

sub Epoll_add :Export(:DEFAULT) {
    croak q{usage: Epoll_add FILEHANDLE, OBJECT|CLASS}  if @_ != 2;
    my ($fh, $disp) = @_;
    croak q{Epoll_add: $_[0] undefined}                 if !defined $fh;
    my $fd = fileno($fh);
    croak q{Epoll_add: $_[0] isn't open FILEHANDLE}     if !defined $fd;
    croak q{Epoll_add: $_[1] undefined}                 if !defined $disp;
    croak q{Epoll_add: $_[1] can't be CODE}             if ref $disp eq 'CODE';
    if (defined $Disp_for[$fd] && !defined $Disp_for[$fd][0]) {
        eval { Epoll_del($fd) };            # destroyed OBJECT
        croak $@ if $@ && $@ !~ /epoll_del: No such file/;
    }
    croak q{Epoll_add: FILEHANDLE wasn't deleted by Epoll_del()}
        if defined $Disp_for[$fd];
    epoll_add($EPFD, $fd, $EVENTMASK);
    $Disp_for[$fd] = [ $disp, $fh ];
    weaken $Disp_for[$fd][0] if ref $disp;  # don't own this OBJECT
    return $fd;
}

sub Epoll_del :Export(:DEFAULT) {
    croak q{usage: Epoll_del EPOLLID}           if @_ != 1;
    my ($fd) = @_;
    croak q{Epoll_del: $_[0] undefined}         if !defined $_[0];
    croak q{Epoll_del: $_[0] isn't EPOLLID}     if $fd !~ /\A \d+ \z/xms;
    croak q{Epoll_del: wrong EPOLLID}           if !$Disp_for[$fd];
    undef $Disp_for[$fd];
    epoll_del($EPFD, $fd);
    return;
}

sub Epoll_wait :Export(:event) {
    croak  q{usage: Epoll_wait TIMEOUT} if @_ != 1;
    my ($timeout) = @_;
    croak qq{Epoll_wait: wrong TIMEOUT $timeout}
        if defined $timeout && $timeout !~ /\A \d+ (?:\.\d*)? \z/xms;
    my ($status, $events) = epoll_wait($EPFD, $timeout);
    return undef if $status==-1;    # EINTR only!
    my (@exceptions, @events);
    EVENT:
    for (@$events) {
        my ($fd, $events) = @$_;
        if ($Disp_for[$fd]) {
            my ($disp, $fh) = @{ $Disp_for[$fd] };
            if (!defined $disp) {
                eval { Epoll_del($fd) };    # destroyed OBJECT
                # Because of packet nature of Epoll_wait() we can't throw
                # exceptions here because this will result in other events
                # lost. So we return happens exceptions as usual events...
                push @exceptions, [\&croak, undef, $@]
                    if $@ && $@ !~ /epoll_del: No such file/;
                next EVENT;
            }
            push @events, [$disp, 'IN',  $fd, $fh]  if $events & EPOLLIN;
            push @events, [$disp, 'OUT', $fd, $fh]  if $events & EPOLLOUT;
            push @events, [$disp, 'HUP', $fd, $fh]  if $events & EPOLLHUP;
            push @events, [$disp, 'ERR', $fd, $fh]  if $events & EPOLLERR;
        }
        # Here we can catch events for:
        # 1) bug in this code (some our fd's don't processed by above code)
        # 2) bug in epoll code (it return some unknown for us fd's)
        else {
            croak qq{Internal error: no dispatcher for events=$events on fd=$fd};
        }
    }
    return $status,
        @exceptions,
        map { ref($_->[0]) and weaken($_->[0]); $_ } @events
        ;
}

1; # Magic true value required at end of module
__END__

=head1 NAME

POWER::Event::Epoll - I/O event notification facility


=head1 VERSION

This document describes POWER::Event::Epoll version 0.2.0


=head1 SYNOPSIS

 # in user code:
 use POWER::Event::Epoll;
 Epoll_add($socket, $client_object);
 $epoll_id = Epoll_add(\*STDIN, 'MySTDIN');
 Epoll_del($epoll_id);

 # somethere in event loop:
 use POWER::Utils;                     # for run_callback()
 use POWER::Event::Epoll qw( :event );
 my ($status, @events) = Epoll_wait();
 if (!defined $status) {
     # signal arrived
 }
 elsif (!$status) {
     # timeout happens
 }
 else {
     run_callback(@$_) for @events;
 }

=head1 DESCRIPTION

Allow you to:

 - set a lot of FILEHANDLEs for monitoring in L<epoll(4)>
   and bind each FILEHANDLE to your OBJECT or CLASS
 - wait until:
    a) I/O events happens on some of monitored FILEHANDLEs
    b) timeout expires
    c) signal arrives
   get list of happens I/O events in format suitable for executing
   predefined methods (IN/OUT/HUP/ERR) of your OBJECT/CLASS using
   POWER::Utils/run_callback()
 - remove your FILEHANDLEs from monitoring

Interface of this module it developed for using in user's event loops,
for example see POWER::Event.


=head1 EXPORTS

=over

=item :DEFAULT

Epoll_add()
Epoll_del()


=item :event

Epoll_wait()


=back

=head1 INTERFACE

=over

=item B<Epoll_add>( FILEHANDLE, OBJECT|CLASS )

Add FILEHANDLE to monitoring. All events on FILEHANDLE will be dispatched
to OBJECT|CLASS methods IN(), OUT(), HUP(), ERR().

If OBJECT given then B<weaken> reference to OBJECT will be stored in
internal structures. If that OBJECT will be destroyed without calling
Epoll_del() first, then Epoll_wait() I<will not return events for OBJECT>.
in case of any events happens on FILEHANDLE for such destroyed OBJECT
Epoll_wait() will automatically call Epoll_del(), so you I<shouldn't>
manually call Epoll_del() for already-destroyed OBJECT's.

In addition Epoll_add() will throw exception if you try to add another
FILEHANDLE with same fileno() without calling Epoll_del() first.

Return: epoll identifier which can be used with Epoll_del().


=item B<Epoll_del>( EPOLLID )

Remove EPOLLID from monitoring.

NOTE: Reference to OBJECT stored by Epoll_add() also will be deleted.

Return: nothing.


=item B<Epoll_wait>( TIMEOUT )

Wait for one of:

 - some I/O events on FILEHANDLEs added by Epoll_add()
 - TIMEOUT expires
 - signal arrives

Return: (STATUS, EVENTS), where:

 STATUS     if signal arrives     : undefined scalar
            if timeout expires    : 0
            if I/O events happens : positive number
 EVENTS     list with all I/O events in format suitable for
            POWER::Utils/run_callback() (see L</SYNOPSIS> for usage
            example)


=back

=head1 DIAGNOSTICS

=head2 EXCEPTIONS

=over

=item C<< usage: Epoll_add FILEHANDLE, OBJECT|CLASS >>

You must call Epoll_add() with two params.


=item C<< Epoll_add: $_[0] undefined >>

=item C<< Epoll_add: $_[0] isn't open FILEHANDLE >>

First param of Epoll_add() must be open FILEHANDLE.


=item C<< Epoll_add: $_[1] undefined >>

=item C<< Epoll_add: $_[1] can't be CODE >>

POWER::Event::Epoll doesn't support CODE callbacks like
POWER::Utils/run_callback() do.
You must provide OBJECT or CLASS name instead.


=item C<< Epoll_add: FILEHANDLE wasn't deleted by Epoll_del() >>

You try to Epoll_add() FILEHANDLE, but this FILEHANDLE has same fileno()
as some previously Epoll_add()'ed FILEHANDLE which wasn't removed by
Epoll_del().


=item C<< usage: Epoll_del EPOLLID >>

Epoll_del() require exactly one param.


=item C<< Epoll_del: $_[0] undefined >>

=item C<< Epoll_del: $_[0] isn't EPOLLID >>

You try to use Epoll_del() with something strange. Epoll_del() must be used
with value returned by Epoll_add().


=item C<< Epoll_del: wrong EPOLLID >>

Epoll_del() unable to delete this EPOLLID. Possible reasons: you already
deleted this EPOLLID with Epoll_del() or EPOLLID isn't a value returned
by Epoll_add().


=item C<< usage: Epoll_wait TIMEOUT >>

Epoll_wait() require exactly one param.


=item C<< Epoll_wait: wrong TIMEOUT $timeout >>

TIMEOUT must be undefined or be a number (with optional fractional part).


=item C<< Internal error: no dispatcher for events=$events on fd=$fd >>

You've find a bug in POWER::Event::Epoll.


=item C<< epoll_*: ... >>

See L<POWER::Epoll> about such low-level errors.


=back

=head2 WARNINGS

None.


=head1 CONFIGURATION AND ENVIRONMENT

POWER::Event::Epoll requires no configuration files or environment variables.


=head1 DEPENDENCIES

 version
 Perl6::Export::Attrs

 POWER::Epoll           2.0.0


=head1 INCOMPATIBILITIES

None reported.


=head1 BUGS AND LIMITATIONS

No bugs have been reported.

Please report any bugs or feature requests to
C<bug-power-event-epoll@rt.cpan.org>, or through the web interface at
L<http://rt.cpan.org>.


=head1 AUTHOR

Alex Efros  C<< <powerman@asdfgroup.com> >>


=head1 LICENCE AND COPYRIGHT

Copyright (c) 2005, Alex Efros C<< <powerman@asdfgroup.com> >>. All rights reserved.

This module is free software; you can redistribute it and/or
modify it under the same terms as Perl itself. See L<perlartistic>.


=head1 DISCLAIMER OF WARRANTY

BECAUSE THIS SOFTWARE IS LICENSED FREE OF CHARGE, THERE IS NO WARRANTY
FOR THE SOFTWARE, TO THE EXTENT PERMITTED BY APPLICABLE LAW. EXCEPT WHEN
OTHERWISE STATED IN WRITING THE COPYRIGHT HOLDERS AND/OR OTHER PARTIES
PROVIDE THE SOFTWARE "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE
ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE SOFTWARE IS WITH
YOU. SHOULD THE SOFTWARE PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL
NECESSARY SERVICING, REPAIR, OR CORRECTION.

IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MAY MODIFY AND/OR
REDISTRIBUTE THE SOFTWARE AS PERMITTED BY THE ABOVE LICENCE, BE
LIABLE TO YOU FOR DAMAGES, INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL,
OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OR INABILITY TO USE
THE SOFTWARE (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING
RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD PARTIES OR A
FAILURE OF THE SOFTWARE TO OPERATE WITH ANY OTHER SOFTWARE), EVEN IF
SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF
SUCH DAMAGES.
