use warnings;
use strict;
use Test::More tests => 32;
use Test::Exception;
use Test::MockModule;
use Errno qw( ENOMEM ENOENT EINVAL );

use POWER::Epoll;
use POWER::Utils;
use POWER::Event::Epoll qw( :ALL );


throws_ok { Epoll_add()                             } qr/usage: Epoll_add/;
throws_ok { Epoll_add(\*STDIN)                      } qr/usage: Epoll_add/;
throws_ok { Epoll_add(\*STDIN, 'MyClass', undef)    } qr/usage: Epoll_add/;
throws_ok { Epoll_add(\*STDIN, undef)               } qr/\$_\[1] undefined/;
throws_ok { Epoll_add(undef, 'MyClass')             } qr/undefined/;
throws_ok { Epoll_add(q{}, 'MyClass')               } qr/isn't open FILEHANDLE/;
throws_ok { Epoll_add(0, 'MyClass')                 } qr/isn't open FILEHANDLE/;
throws_ok { Epoll_add(*STDIN, sub{})                } qr/can't be CODE/;

# Epoll_add() without Epoll_del() on destroyed OBJECT, without error
{
    my $module = Test::MockModule->new('POWER::Event::Epoll');
    $module->mock('epoll_add' => sub { 1 });
    my $is_deleted = 0;
    $module->mock('epoll_del' => sub { $is_deleted++ });
    my $obj = bless {}, 'MyClass';
    Epoll_add(*STDIN, $obj);
    undef $obj; 
    my $stdin;
    lives_ok  { $stdin = Epoll_add(*STDIN, 'MyClass') } q{automaticall Epoll_del()};
    ok($is_deleted, 'epoll_del really was executed');
    $module->mock('epoll_del' => sub { 1 });
    Epoll_del($stdin);
}

# Epoll_wait() without Epoll_del() on destroyed OBJECT, with error
{
    my $module = Test::MockModule->new('POWER::Event::Epoll');
    $module->mock('epoll_add' => sub { 1 });
    $module->mock('epoll_del' => sub { die 'epoll_del: something' });
    $module->mock('epoll_wait' => sub { return (1, [
        [ fileno(*STDIN), EPOLLIN ],
    ]) });
    my $obj = bless {}, 'MyClass';
    my $stdin = Epoll_add(*STDIN, $obj);
    undef $obj; 
    my ($status, @events) = Epoll_wait(0);
    throws_ok { run_callback(@$_) for @events } qr/epoll_del: something/;
}

throws_ok { Epoll_del()                             } qr/usage: Epoll_del/;
throws_ok { Epoll_del(0, 0)                         } qr/usage: Epoll_del/;
throws_ok { Epoll_del(undef)                        } qr/undefined/;
throws_ok { Epoll_del(\*STDIN)                      } qr/isn't EPOLLID/;
throws_ok { Epoll_del(255)                          } qr/wrong EPOLLID/;

my $epoll1;
lives_ok  { $epoll1 = Epoll_add(*STDIN, 'MyClass')  } q{STDIN -> MyClass};
throws_ok { Epoll_add(*STDIN, 'MyClass')            } qr/wasn't deleted/;
lives_ok  { Epoll_del($epoll1)                      } q{STDIN removed};
throws_ok { Epoll_del($epoll1)                      } qr/wrong EPOLLID/;

{
    my $module = Test::MockModule->new('POWER::Event::Epoll');
    $module->mock('epoll_del' => sub { die 'epoll_del: something' });
    $epoll1 = Epoll_add(*STDIN, 'MyClass');
    throws_ok { Epoll_del($epoll1)                  } qr/epoll_del: something/;
}
throws_ok { Epoll_del($epoll1)                      } qr/wrong EPOLLID/;

throws_ok { Epoll_wait()                            } qr/usage: Epoll_wait/;
throws_ok { Epoll_wait(1,1)                         } qr/usage: Epoll_wait/;
throws_ok { Epoll_wait(q{})                         } qr/wrong TIMEOUT/;
throws_ok { Epoll_wait(-1)                          } qr/wrong TIMEOUT/;
throws_ok { Epoll_wait(q{qwe})                      } qr/wrong TIMEOUT/;
throws_ok { Epoll_wait([])                          } qr/wrong TIMEOUT/;
{
    my $module = Test::MockModule->new('POWER::Event::Epoll');
    $module->mock('epoll_wait' => sub { return(0, []) });
    lives_ok { Epoll_wait(undef)                    } q{epoll_wait(undef)};
    lives_ok { Epoll_wait(0)                        } q{epoll_wait(0)};
    lives_ok { Epoll_wait(0.1)                      } q{epoll_wait(0.1)};
}

{
    my $module = Test::MockModule->new('POWER::Event::Epoll');
    $module->mock('epoll_wait' => sub { $!=EINVAL; return(1, [[256,0]]) });
    throws_ok { Epoll_wait(0)                       } qr/Internal error: no dispatcher/;
}

